from typing import TYPE_CHECKING, Any, List, TypeVar

from laminar.configurations.hooks import annotation
from laminar.types import hints

if TYPE_CHECKING:
    from laminar.components import Layer

T = TypeVar("T")


def entry(hook: T) -> T:
    """Configure an entry hook

    Usage::

        from laminar.configurations import entry

        @hooks.entry
        def enter() -> bool:
            ...
    """

    return annotation.annotate(hook, annotation.entry)


def gather(*, layer: "Layer", annotation: str) -> List[Any]:
    """Get values returned by all hooks of the annotated type.

    Args:
        layer: Layer the hooks are for.
        annotation: Annotation to get the hooks for.

    Returns:
        Return values for each hook.
    """

    return [hook(layer, *hints(layer.flow, hook)) for hook in layer.hooks.get(annotation, [])]



"""Unit tests for laminar.configurations.protocol"""

from laminar.configurations import serde


class TestDtype:
    def test_builtin(self) -> None:
        assert serde.dtype(str) == "builtins.str"

    def test_custom(self) -> None:
        assert serde.dtype(serde.Protocol) == "laminar.configurations.serde.Protocol"



import inspect
from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING, TypeVar

from laminar.configurations.hooks import annotation
from laminar.types import hints

if TYPE_CHECKING:
    from laminar.components import Layer

T = TypeVar("T")


def execution(hook: T) -> T:
    """Configure an execution hook.

    Usage::

        from laminar.configurations import hooks

        @hooks.execution
        def configure() -> Generator[None, None, None]:
            ...
    """

    return annotation.annotate(hook, annotation.execution)


def retry(hook: T) -> T:
    """Configure a retry hook.

    Usage::

        from laminar.configurations import hooks

        @hooks.retry
        def configure() -> Generator[None, None, None]:
            ...
    """

    return annotation.annotate(hook, annotation.retry)


def submission(hook: T) -> T:
    """Configure a submission hook.

    Usage::

        from laminar.configurations import hooks

        @hooks.submission
        def configure() -> Generator[None, None, None]:
            ...
    """

    return annotation.annotate(hook, annotation.submission)


def context(*, layer: "Layer", annotation: str) -> ExitStack:
    """Get a context manager and results for all hooks of the annotated type.

    Args:
        layer: Layer the hooks are for.
        annotation: Annotation to get hooks for.

    Returns:
        A context manager with all annotated hooks activated.
    """

    stack = ExitStack()
    for hook in layer.hooks.get(annotation, []):
        # Gather any layer dependencies the hook may have
        parameters = hints(layer.flow, hook)

        # Create a context manager for the generator and register it with the exit stack
        if inspect.isgeneratorfunction(hook):
            manager = contextmanager(hook)
            stack.enter_context(manager(layer, *parameters))

        # Call the hook function
        else:
            hook(layer, *parameters)

    return stack



"""Shared laminar utilities."""


def stringify(obj: object, name: str, *attrs: str) -> str:
    """Create a string representation of an object.

    Args:
        obj: Object the representation is for.
        name: Name of the object
        *attrs: Attributes to include in the representation.

    Returns:
        Object's string representation.
    """

    variables = {attr: getattr(obj, attr, None) for attr in attrs} if attrs else vars(obj)
    attributes = ", ".join(f"{key}={repr(value)}" for key, value in variables.items())
    return f"{name}({attributes})"



"""Shared laminar types."""

from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Type, TypeVar, get_type_hints

if TYPE_CHECKING:
    from laminar import Flow, Layer
else:
    Flow = "Flow"
    Layer = "Layer"

T = TypeVar("T")
LayerType = TypeVar("LayerType", bound=Type[Layer])


def hints(flow: Flow, function: Callable[..., Any]) -> Tuple[Layer, ...]:
    """Get the type hints for a given function.

    Args:
        flow: Flow to get layers for.
        function (Callable[..., Any]): Function to get type hints for.

    Returns:
        Ordered type hints.
    """

    return tuple(
        flow.layer(hint)
        for hint in (annotation for parameter, annotation in get_type_hints(function).items() if parameter != "return")
    )


def unwrap(option: Optional[T], default: Optional[T] = None) -> T:
    """Unwrap an optional value.

    Usage::

        unwrap("a")
        >>> "a"
        unwrap(None, "a")
        >>> "a"
        unwrap(None)
        >>> ValueError

    Args:
        option: Optional value to unwrap.
        default: Default value if the optional value is None.

    Raises:
        ValueError: If the option is None and no default is set.

    Returns:
        Unwrapped optional value.
    """

    if option is None:
        if default is None:
            raise ValueError("Value is None when it shouldn't be.")
        return unwrap(default)
    return option



"""Test composed flows."""

import pytest

from laminar import Flow, Layer
from laminar.components import Parameters
from laminar.configurations import datastores, executors
from laminar.types import unwrap


class Flow1(Flow):
    ...


@Flow1.register()
class A(Layer):
    def __call__(self) -> None:
        self.foo = "bar"


class Flow2(Flow):
    ...


@Flow2.register()
class B(Layer):
    def __call__(self, parameters: Parameters) -> None:
        self.foo = parameters.foo


class Flow3(Flow):
    ...


@Flow3.register()
class C(Layer):
    def __call__(self, parameters: Parameters) -> None:
        self.foo = parameters.foo


@pytest.mark.flow
def test_flow() -> None:
    flow1 = Flow1(datastore=datastores.Memory(), executor=executors.Thread())
    flow2 = Flow2(datastore=datastores.Memory(), executor=executors.Thread())
    flow3 = Flow3(datastore=datastores.Memory(), executor=executors.Thread())

    execution = (
        flow1()
        .chain(flow=flow2, linker=lambda execution: Parameters(foo=execution.layer(A).foo))
        .chain(flow=flow3, linker=lambda execution: Parameters(foo=execution.layer(B).foo))
    )

    assert flow2.execution(unwrap(execution.id)).layer(B).foo == "bar"
    assert flow3.execution(unwrap(execution.id)).layer(C).foo == "bar"


class CombinedFlow(Flow1, Flow2, Flow3):
    ...


@pytest.mark.flow
def test_combination() -> None:
    assert sorted(CombinedFlow.registry) == ["A", "B", "C", "Parameters"]
    assert CombinedFlow.registry["A"] == A()
    assert CombinedFlow.registry["B"] == B()
    assert CombinedFlow.registry["C"] == C()

    flow = CombinedFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow(foo="bar")

    assert flow.execution(unwrap(execution.id)).layer(B).foo == "bar"
    assert flow.execution(unwrap(execution.id)).layer(C).foo == "bar"



from laminar.configurations.hooks import annotation, condition, event
from laminar.configurations.hooks.condition import entry
from laminar.configurations.hooks.event import execution, retry, submission

__all__ = ["annotation", "condition", "entry", "event", "execution", "retry", "submission"]



"""Unit tests for laminar.utils.fs"""

from pathlib import Path

from laminar.utils import fs


def test_exists() -> None:
    assert fs.exists(uri=str(Path(__file__).resolve()))
    assert not fs.exists(uri=str(Path(__file__).resolve() / "foo.bar"))


def test_join() -> None:
    assert fs.join("path", "to", "something") == "path/to/something"



"""Configurations for laminar layers."""

import asyncio
import itertools
import logging
import random
import sys
from dataclasses import dataclass, field
from inspect import Traceback
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, Type

from laminar.configurations import datastores
from laminar.settings import current
from laminar.types import unwrap

if TYPE_CHECKING:
    from laminar import Layer
else:
    Layer = "Layer"

logger = logging.getLogger(__name__)


@dataclass
class Catch:
    """Configures a layer to catch certain exceptions.

    Notes:

        Catches specified errors and their subclasses.

    Usage::

        @Flow.register(catch=Catch(...))
    """

    def __init__(self, *exceptions: Type[Exception]) -> None:
        self.exceptions = exceptions

    def __enter__(self) -> "Catch":
        return self

    def __exit__(self, etype: Type[Exception], error: Exception, traceback: Traceback) -> Optional[Type[Exception]]:
        if isinstance(error, self.exceptions):
            self.exception = etype
            return etype
        return None


@dataclass
class Container:
    """Configures a layer's container properties.

    Notes:

        Most configurations map one-to-one with a configuration here:
        https://docs.docker.com/engine/reference/commandline/container_run/

    Usage::

        @Flow.register(container=Container(...))
    """

    #: Command to execute in the container
    command: str = "python main.py"
    #: vCPUs to allocate to the container
    cpu: int = 1
    #: Image to create the container with
    image: str = f"rchui/laminar:{'.'.join(map(str, sys.version_info[:2]))}"
    #: Memory in megabytes
    memory: int = 1500
    #: Directory to execute the command in
    workdir: str = "/laminar"

    def __post_init__(self) -> None:
        if not self.workdir.endswith(("://", ":///")):
            object.__setattr__(self, "workdir", self.workdir.rstrip("/"))


@dataclass
class Parameter:
    """Input for configuring a ForEach."""

    #: Layer the attribute is associated with.
    layer: Type["Layer"]
    #: Attribute to iterate over.
    attribute: str
    #: Layer index to reference attributes from.
    index: Optional[int] = 0


@dataclass
class ForEach:
    """Configures a layer to perform a grid-foreach over the configured properties.

    Notes:

        In order to foreach over each element of a layer, the layer's artifact must be sharded. See Layer.shard().

        If `Parameter(index=None)`, ForEach will include artifacts from all Layer splits.

    Usage::

        @Flow.register(foreach=ForEach(...))
    """

    parameters: Iterable[Parameter] = field(default_factory=list)  #: Parameters to configure the foreach with.

    def join(self, *, layer: Layer, name: str) -> datastores.Archive:
        """Join together multiple artifact splits of a layer into a single Archive.

        Args:
            layer (Layer): Layer to join archives for.
            name (str): Name of the attribute to join archives for.

        Returns:
            datastores.Archive: Archive created from multiple layer archives.
        """

        datastore = layer.flow.configuration.datastore

        if datastore.exists(path=datastores.Archive.path(layer=layer, index=0, name=name, cache=True)):
            logger.debug("Cache hit for layer '%s', archive '%s'.", layer.name, name)
            archive = datastore.read_archive(layer=layer, index=0, name=name, cache=True)

        else:
            logger.debug("Cache miss for layer '%s', archive '%s'.", layer.name, name)
            archives = [
                datastore.read_archive(layer=layer, index=index, name=name)
                for index in range(layer.configuration.foreach.splits(layer=layer))
            ]
            artifacts = list(itertools.chain.from_iterable([archive.artifacts for archive in archives]))

            archive = datastore.write_archive(layer=layer, name=name, artifacts=artifacts, cache=True)

        return archive

    def grid(self, *, layer: Layer) -> List[Dict[Layer, Dict[str, int]]]:
        """Generate a grid of all combinations of foreach inputs.

        Args:
            layer (Layer): Layer the ForEach is configured for.

        Returns:
            List[Dict[Layer, Dict[str, int]]]: Index ordered inputs of layers mapped to attributes mapped to Accessor
                index.
        """

        parameters: List[Tuple[Layer, str]] = []
        archives: List[datastores.Archive] = []

        for parameter in self.parameters:
            instance = layer.flow.layer(parameter.layer)
            parameters.append((instance, parameter.attribute))

            # Get archives for all layer splits.
            if parameter.index is None:
                archives.append(instance.configuration.foreach.join(layer=instance, name=parameter.attribute))

            # Get archive for specified layer index.
            else:
                archives.append(
                    instance.flow.configuration.datastore.read_archive(
                        layer=instance, index=parameter.index, name=parameter.attribute
                    )
                )

        # Compute the product of every possible set of parameter indexes based off of parameter layer splits.
        grid: List[Dict[Layer, Dict[str, int]]] = []
        for indexes in itertools.product(*(range(len(archive)) for archive in archives)):
            model: Dict[Layer, Dict[str, int]] = {}
            for (instance, attribute), index in zip(parameters, indexes):
                model.setdefault(instance, {})[attribute] = index
            grid.append(model)

        return grid

    def splits(self, *, layer: Layer) -> int:
        """Get the splits of the ForEach grid."""

        if layer.flow.configuration.datastore.exists(path=datastores.Record.path(layer=layer)):
            logger.debug("Cache hit for layer '%s' record.", layer.name)
            return layer.flow.configuration.datastore.read_record(layer=layer).execution.splits

        else:
            logger.debug("Cache miss for layer '%s' record.", layer.name)
            return len(self.grid(layer=layer))

    def set(self, *, layer: Layer, parameters: Tuple[Layer, ...]) -> Tuple[Layer, ...]:
        """Set a foreach layer's parameters given the inputs from the foreach grid.

        Args:
            layer (Layer): Layer the ForEach is configured for.
            parameters (Tuple[Layer, ...]): Parameters in the order they will be passed to the layer.

        Returns:
            Tuple[Layer, ...]: Parameters with modified values for the foreach evaluation.
        """

        inputs = self.grid(layer=layer)[unwrap(layer.index)]
        for parameter in parameters:
            for attribute, index in inputs.get(parameter, {}).items():
                value = getattr(parameter, attribute)

                # Index only if the requested attribute is an Accessor
                value = value[index] if isinstance(value, datastores.Accessor) else value
                setattr(parameter, attribute, value)

        return parameters


@dataclass
class Retry:
    """Configure a Layer to retry on failure using exponential backoff.

    Notes:

        Computes the retry backoff with:

            >>> sleep = <delay> * (<backoff> ** (attempt - 1))
            >>> sleep = sleep + random(0, sleep * <jitter>)

    Usage::

        @Flow.register(retry=Retry(...))
    """

    #: Number of retires to attempt before failing
    attempts: int = 1
    #: Base number of seconds to wait before retrying
    delay: float = 0.1
    #: Backoff factor to multiply delay with
    backoff: float = 2.0
    #: Factor to randomize delay.
    jitter: float = 0.1

    async def sleep(self, *, layer: Layer, attempt: int) -> None:
        """Exponentially backoff before retrying a layer.

        Args:
            layer: Layer being retried.
            attempt: Attempt the layer is on.
        """

        sleep = layer.configuration.retry.delay * (layer.configuration.retry.backoff ** (attempt - 1))
        sleep = sleep + random.uniform(0, sleep * layer.configuration.retry.jitter)
        sleep = round(sleep, 2)

        logger.info("Retrying layer '%s' after backing off for '%.2f' seconds.", layer.name, sleep)

        await asyncio.sleep(sleep)


@dataclass
class Configuration:
    """Layer configurations.

    Usage::

        class A(Layer):
            def __call__(self) -> None:
                self.configuration.container
                self.configuration.foreach
                self.configuration.retry
    """

    catch: Catch = Catch()
    #: Layer container configuration
    container: Container = Container()
    #: Layer foreach configuration
    foreach: ForEach = ForEach()
    #: Layer retry configuration
    retry: Retry = Retry()


@dataclass
class State:
    layer: Layer

    @property
    def finished(self) -> bool:
        return self.layer.flow.configuration.datastore.exists(
            path=datastores.Record.path(layer=self.layer.flow.layer(self.layer))
        )

    @property
    def running(self) -> bool:
        return current.layer.name is not None and current.layer.name == self.layer.name



"""Unit tests for laminar.configurations.executors"""

import asyncio
import shlex
from unittest.mock import Mock, patch

import pytest

from laminar import Layer
from laminar.configurations.executors import Docker, Executor, Thread
from laminar.exceptions import ExecutionError


@pytest.mark.asyncio
class TestExecutor:
    executor = Executor()

    async def test_sempahore(self) -> None:
        assert self.executor.semaphore._value == 1
        assert self.executor.semaphore != asyncio.Semaphore(1)

    async def test_submit(self, layer: Layer) -> None:
        with pytest.raises(NotImplementedError):
            await self.executor.submit(layer=layer)


@pytest.mark.asyncio
class TestThread:
    executor = Thread()

    async def test_submit(self) -> None:
        mock_layer = Mock()

        assert await self.executor.submit(layer=mock_layer) == mock_layer

        mock_layer.flow.execute.assert_called_once_with(execution=mock_layer.flow.execution.id, layer=mock_layer)


@pytest.mark.asyncio
class TestDocker:
    executor = Docker()

    async def test_submit(self, layer: Layer) -> None:
        command = shlex.split("echo 'hello world'")
        with patch("shlex.split") as mock_split:
            mock_split.return_value = command

            assert await self.executor.submit(layer=layer) == layer

        mock_split.assert_called_once_with(
            "docker run --rm --interactive --cpus 1 --env LAMINAR_EXECUTION_ID=test-execution --env"
            " LAMINAR_EXECUTION_RETRY=False --env LAMINAR_FLOW_NAME=TestFlow --env LAMINAR_LAYER_ATTEMPT=1 --env"
            " LAMINAR_LAYER_INDEX=0 --env LAMINAR_LAYER_NAME=Layer --env LAMINAR_LAYER_SPLITS=2 --memory 1500m"
            f" --volume memory:///:/laminar/.laminar --workdir /laminar {layer.configuration.container.image} python"
            " main.py"
        )

    async def test_submit_error_code(self, layer: Layer) -> None:
        command = shlex.split("exit 1")
        with patch("shlex.split") as mock_split:
            mock_split.return_value = command

            with pytest.raises(ExecutionError):
                await self.executor.submit(layer=layer)

    async def test_submit_exeception(self, layer: Layer) -> None:
        with patch("shlex.split") as mock_split:
            mock_split.side_effect = Mock(side_effect=Exception())

            with pytest.raises(Exception):
                await self.executor.submit(layer=layer)



import copy
import operator
from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, TypeVar, Union, overload

from laminar.configurations.datastores import DataStore, Local
from laminar.configurations.executors import Docker, Executor
from laminar.configurations.schedulers import Scheduler
from laminar.settings import current
from laminar.types import unwrap
from laminar.utils import stringify

if TYPE_CHECKING:
    from laminar import Flow, Layer
    from laminar.components import Parameters

T = TypeVar("T", bound="Layer")


@dataclass(frozen=True)
class Configuration:
    """Flow configurations.

    Usage::

        class A(Layer):
            def __call__(self) -> None:
                self.flow.configuration.datastore
                self.flow.configuration.executor
                self.flow.configuration.scheduler
    """

    #: Flow datastore configuration
    datastore: DataStore = Local()
    #: Flow executor configuration
    executor: Executor = Docker()
    #: Flow scheduler configuration
    scheduler: Scheduler = Scheduler()


@dataclass
class Execution:
    #: ID of the flow execution
    id: Optional[str]
    #: Flow being executed
    flow: "Flow"
    #: True if the flow execution is being retried, else False.
    retry: bool = False

    def __call__(self, id: str) -> "Execution":
        execution = copy.deepcopy(self)
        execution.id = id
        return execution

    def __repr__(self) -> str:
        return stringify(self, type(self).__name__, "id", "retry")

    @property
    def finished(self) -> bool:
        return reduce(operator.and_, [layer.state.finished for layer in self.flow._dependencies.keys()])

    @property
    def running(self) -> bool:
        return (current.execution.id is not None and current.execution.id == self.id) and (
            current.flow.name is not None and current.flow.name == self.flow.name
        )

    def chain(self, *, flow: "Flow", linker: Callable[["Execution"], "Parameters"]) -> "Execution":
        """Chain multiple flow executions together.

        Usage::

            class Flow1(Flow):
                ...

            class Flow2(Flow):
                ...

            @Flow1.register()
            class A(Layer):
                foo: str

            Flow1()().chain(Flow2(), lambda execution: Parameters(foo=execution.layer(A).foo))

        Args:
            flow: Flow to execute next.
            linker: Function for passing parameters to the next flow.
        """

        artifacts = (
            linker(self).artifacts
            if all(variable is None for variable in (current.execution.id, current.flow.name, current.layer.name))
            else {}
        )
        return flow(execution=self.id, **artifacts)

    @overload
    def layer(self, layer: str, **atributes: Any) -> "Layer":
        ...

    @overload
    def layer(self, layer: Type[T], **attributes: Any) -> T:
        ...

    @overload
    def layer(self, layer: T, **attributes: Any) -> T:
        ...

    def layer(self, layer: Union[str, Type["Layer"], "Layer"], **attributes: Any) -> "Layer":
        """Get a registered flow layer.

        Usage::

            flow.execution(...).layer("A")
            flow.execution(...).layer(A)
            flow.execution(...).layer(A())
            flow.execution(...).layer(A(), index=0, splits=2)

        Args:
            layer: Layer to get.
            **attributes: Keyword attributes to add to the Layer.

        Returns:
            Layer that is registered to the flow.
        """

        return self.flow.layer(layer, **attributes)

    def resume(self) -> None:
        """Resume a flow execution from where it failed.

        Notes:

            Resuming a flow execution will skip all layers that finished on the previous attempt.

        Usage::

            flow.execution(...).resume()
        """

        self.retry = True
        self.flow.schedule(execution=unwrap(self.id), dependencies=self.flow.dependencies)



"""Configurations for laminar executors."""

import asyncio
import hashlib
import logging
import shlex
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

import boto3
from mypy_boto3_batch.client import BatchClient
from mypy_boto3_batch.type_defs import (
    ContainerOverridesTypeDef,
    ContainerPropertiesTypeDef,
    JobTimeoutTypeDef,
    KeyValuePairTypeDef,
)

from laminar.exceptions import ExecutionError
from laminar.types import unwrap
from laminar.utils import contexts

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from laminar import Layer
else:
    Layer = "Layer"


@dataclass(frozen=True)
class Executor:
    """Generic base executor."""

    #: Number of tasks that can execute concurrently.
    concurrency: int = 1
    #: Number of seconds to wait before automatically failing.
    timeout: int = 86400

    @property
    def semaphore(self) -> asyncio.Semaphore:
        """Create a semaphore that limits asyncio concurrency.

        Notes:

            Concurrency is controled by Executor.concurrency

        Usage::

            async with self.semaphore:
                ...
        """

        attr = "_semaphore"
        if not hasattr(self, attr):
            object.__setattr__(self, attr, asyncio.Semaphore(self.concurrency))
        semaphore: asyncio.Semaphore = getattr(self, attr)
        return semaphore

    async def submit(self, *, layer: Layer) -> Layer:
        """Execute a layer.

        Args:
            layer: Layer to execute.
        """

        raise NotImplementedError


@dataclass(frozen=True)
class Thread(Executor):
    """Execute layers in threads.

    Usage::

        Flow(executor=Thread())
    """

    async def submit(self, *, layer: Layer) -> Layer:
        async with self.semaphore:
            with contexts.Environment(
                LAMINAR_EXECUTION_ID=layer.flow.execution.id,
                LAMINAR_EXECUTION_RETRY=layer.flow.execution.retry,
                LAMINAR_FLOW_NAME=layer.flow.name,
                LAMINAR_LAYER_ATTEMPT=layer.attempt,
                LAMINAR_LAYER_INDEX=layer.index,
                LAMINAR_LAYER_NAME=layer.name,
                LAMINAR_LAYER_SPLITS=layer.splits,
            ):
                layer.flow.execute(execution=unwrap(layer.flow.execution.id), layer=layer)

            return layer


@dataclass(frozen=True)
class Docker(Executor):
    """Execute layers in Docker containers.

    Usage::

        Flow(executor=Docker())
    """

    async def submit(self, *, layer: Layer) -> Layer:
        async with self.semaphore:
            workspace = f"{layer.flow.configuration.datastore.root}:{layer.configuration.container.workdir}/.laminar"

            command = " ".join(
                [
                    "docker",
                    "run",
                    "--rm",
                    "--interactive",
                    f"--cpus {layer.configuration.container.cpu}",
                    f"--env LAMINAR_EXECUTION_ID={unwrap(layer.flow.execution.id)}",
                    f"--env LAMINAR_EXECUTION_RETRY={layer.flow.execution.retry}",
                    f"--env LAMINAR_FLOW_NAME={layer.flow.name}",
                    f"--env LAMINAR_LAYER_ATTEMPT={unwrap(layer.attempt)}",
                    f"--env LAMINAR_LAYER_INDEX={unwrap(layer.index)}",
                    f"--env LAMINAR_LAYER_NAME={layer.name}",
                    f"--env LAMINAR_LAYER_SPLITS={unwrap(layer.splits)}",
                    f"--memory {layer.configuration.container.memory}m",
                    f"--volume {workspace}",
                    f"--workdir {layer.configuration.container.workdir}",
                    layer.configuration.container.image,
                    layer.configuration.container.command,
                ]
            )
            logger.debug(command)

            task = asyncio.create_task(asyncio.create_subprocess_exec(*shlex.split(command)))

            try:
                process = await asyncio.wait_for(task, timeout=self.timeout)
                if await process.wait() != 0:
                    raise ExecutionError(f"Layer '{layer.name}' failed with exit code: {process.returncode}")
                return layer
            except ExecutionError:
                raise
            except asyncio.TimeoutError as error:
                raise ExecutionError(f"Layer '{layer.name}' timed out after '{self.timeout}' seconds.") from error
            except Exception as error:
                message = type(error).__name__ + (f":{error}" if str(error) else "")
                raise ExecutionError(f"Layer '{layer.name}' failed with an unexpected error. {message}") from error


class AWS:
    """Execute layers in AWS."""

    @dataclass(frozen=True)
    class BaseBatch:
        #: Amazon resource name (ARN) of an Batch job queue.
        job_queue_arn: str
        #: Amazon resource name (ARN) of an IAM role to attach to each Batch job.
        job_role_arn: str
        #: Poll interval wait between requesting a Batch job's status.
        poll: float = 30.0

    @dataclass(frozen=True)
    class Batch(Executor, BaseBatch):
        """Execute layers in AWS Batch.

        Usage::

            Flow(executor=AWS.Batch())
        """

        async def wait(self, *, layer: Layer, job: str, batch: Optional[BatchClient] = None) -> Layer:
            """Wait on the completion of a Layer.

            Args:
                layer: Layer to wait on
                job: ID of the job the layer was submitted in
                batch: AWS Batch client

            Raises:
                ExecutionError: If the Layer does not finish successfully.
            """

            batch = batch or boto3.client("batch")

            while True:
                job_response = batch.describe_jobs(jobs=[job])
                status = job_response["jobs"][-1]["status"]

                if status == "SUCCEEDED":
                    return layer
                elif status == "FAILED":
                    raise ExecutionError(
                        f"Layer '{layer.name}' failed with status: {job_response['jobs'][0]['statusReason']}"
                    )
                else:
                    logger.info(f"Layer '{layer.name}' has status: {status}")

                await asyncio.sleep(self.poll)

        async def submit(self, *, layer: Layer, batch: Optional[BatchClient] = None) -> Layer:
            async with self.semaphore:
                container = layer.configuration.container

                job_definition_hexdigest = hashlib.sha256((container.image + self.job_role_arn).encode()).hexdigest()
                job_definition_name = f"laminar_{job_definition_hexdigest}"

                batch = batch or boto3.client("batch")
                describe_response = batch.describe_job_definitions(jobDefinitionName=job_definition_name)

                # Job definition exists. Use existing one
                if describe_response["jobDefinitions"]:
                    job_definition_arn = describe_response["jobDefinitions"][-1]["jobDefinitionArn"]

                # Job definition doesn't exist. Create one.
                else:
                    logger.info("Creating job definition '%s'.", job_definition_name)
                    register_response = batch.register_job_definition(
                        jobDefinitionName=job_definition_name,
                        type="container",
                        containerProperties=ContainerPropertiesTypeDef(
                            image=container.image, jobRoleArn=self.job_role_arn
                        ),
                    )
                    job_definition_arn = register_response["jobDefinitionArn"]

                # Submit job to Batch
                submit_response = batch.submit_job(
                    jobName=f"{layer.flow.name}-{layer.flow.execution.id}-{layer.name}-{layer.index}",
                    jobDefinition=job_definition_arn,
                    jobQueue=self.job_queue_arn,
                    containerOverrides=ContainerOverridesTypeDef(
                        vcpus=container.cpu,
                        memory=container.memory,
                        command=container.command,
                        environment=[
                            KeyValuePairTypeDef(name="LAMINAR_EXECUTION_ID", value=unwrap(layer.flow.execution.id)),
                            KeyValuePairTypeDef(name="LAMINAR_EXECUTION_RETRY", value=str(layer.flow.execution.retry)),
                            KeyValuePairTypeDef(name="LAMINAR_FLOW_NAME", value=layer.flow.name),
                            KeyValuePairTypeDef(name="LAMINAR_LAYER_ATTEMPT", value=str(unwrap(layer.attempt))),
                            KeyValuePairTypeDef(name="LAMINAR_LAYER_INDEX", value=str(unwrap(layer.index))),
                            KeyValuePairTypeDef(name="LAMINAR_LAYER_NAME", value=layer.name),
                            KeyValuePairTypeDef(name="LAMINAR_LAYER_SPLITS", value=str(unwrap(layer.splits))),
                        ],
                    ),
                    timeout=JobTimeoutTypeDef(attemptDurationSeconds=self.timeout),
                )
                job_id = submit_response["jobId"]

                # Poll for job completion
                task = asyncio.create_task(self.wait(layer=layer, job=job_id, batch=batch))
                return await asyncio.wait_for(task, timeout=self.timeout)



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Any, BinaryIO, TextIO, Union, overload

import smart_open
from typing_extensions import Literal

parse_uri = smart_open.parse_uri


@overload
def open(uri: str, mode: Literal["r"]) -> TextIO:
    ...


@overload
def open(uri: str, mode: Literal["rb"]) -> BinaryIO:
    ...


@overload
def open(uri: str, mode: Literal["w"]) -> TextIO:
    ...


@overload
def open(uri: str, mode: Literal["wb"]) -> BinaryIO:
    ...


@contextmanager  # type: ignore
def open(uri: str, mode: Literal["r", "rb", "w", "wb"]) -> Union[BinaryIO, TextIO]:  # type: ignore
    """Open a file handler to a local or remote file.

    Usage::

        with fs.open("file:///...", "r") as file: ...
        with fs.open("s3://...", "wb") as file: ...

    Args:
        uri: URI to the file to open.
        mode: Mode to open the file with.

    Returns:
        Union[BinaryIO, TextIO]: File handle to the local/remote file.
    """

    if parse_uri(uri).scheme == "file" and "w" in mode:
        Path(uri).parent.mkdir(parents=True, exist_ok=True)

    with smart_open.open(uri, mode) as file:
        yield file


def exists(*, uri: str) -> bool:
    """Check for the existance of a local/remote file.

    Usage::

        fs.exists("file:///...")
        fs.exists("s3://...")

    Args:
        uri: URI to the file to check.

    Returns:
        bool: True if the file exists, else False.
    """

    try:
        with open(uri, "rb"):
            return True
    except IOError:
        return False


def join(*parts: Any) -> str:
    """Join parts into a path

    Args:
        *parts: Parts to use to construct the URI.

    Returns:
        Parts joined together
    """

    return os.path.join(*map(str, parts))



"""Configurations for laminar serde."""

from typing import Any, BinaryIO, TypeVar

import cloudpickle

from laminar.utils import fs, stringify


def dtype(cls: type) -> str:
    """Get the serde dtype name given a class type."""

    return f"{cls.__module__}.{cls.__name__}"


class MetaProtocol(type):
    @property
    def dtype(cls) -> str:
        return dtype(cls)


class Protocol(metaclass=MetaProtocol):
    """Generic base class for defining ser(de) protocols."""

    def __repr__(self) -> str:
        return stringify(self, type(self).__name__)

    def read(self, uri: str) -> Any:
        """Read a value from a URI with a custom protocol.

        Args:
            uri: URI to read from.

        Returns:
            Value from the URI.
        """

        with fs.open(uri, "rb") as file:
            return self.load(file)

    def load(self, file: BinaryIO) -> Any:
        """Deserialize a value from a file.

        Usage::

            with open(..., "rb") as file:
                Protocol().load(file)

        Args:
            file: File handler to read from.

        Returns:
            Deserialized value.
        """

        return self.loads(file.read())

    def loads(self, stream: bytes) -> Any:
        """Deserialize a value from a byte stream.

        Usage::

            Protocol().loads(...)

        Args:
            stream: Bytes to deserialize

        Returns:
            Deserialized value.
        """

        raise NotImplementedError

    def write(self, value: Any, uri: str) -> None:
        """Write a value to a URI with a custom protocol.

        Args:
            value: Value to write to the URI.
            uri: URI to write to.
        """

        with fs.open(uri, "wb") as file:
            self.dump(value, file)

    def dump(self, value: Any, file: BinaryIO) -> None:
        """Serialize a value to a file.

        Usage::

            with open(..., "wb") as file:
                Protocol().dump(..., file)

        Args:
            value: Value to serialize.
            file: File handler to write to.
        """

        file.write(self.dumps(value))

    def dumps(self, value: Any) -> bytes:
        """Serialize a value to a byte string.

        Usage::

            Protocol.dumps(...)

        Args:
            value: Value to serialize.

        Returns:
            Serialized value.
        """

        raise NotImplementedError


ProtocolType = TypeVar("ProtocolType", bound=Protocol)


class PickleProtocol(Protocol):
    """Custom protocol for serializing Python objects using pickle."""

    def load(self, file: BinaryIO) -> Any:
        return cloudpickle.load(file)

    def loads(self, stream: bytes) -> Any:
        return cloudpickle.loads(stream)

    def dump(self, value: Any, file: BinaryIO) -> None:
        cloudpickle.dump(value, file)

    def dumps(self, value: Any) -> bytes:
        stream: bytes = cloudpickle.dumps(value)
        return stream



"""Test resuming flows."""

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors


class ResumeFlow(Flow):
    ...


@ResumeFlow.register()
class A(Layer):
    def __call__(self) -> None:
        ...


@ResumeFlow.register()
class B(Layer):
    fail: bool = True

    def __call__(self, a: A) -> None:
        if self.fail:
            raise RuntimeError


@ResumeFlow.register()
class C(Layer):
    def __call__(self, b: B) -> None:
        ...


@pytest.mark.flow
def test_flow() -> None:
    flow = ResumeFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution_id = "test-execution"

    # Catch failure
    try:
        flow(execution=execution_id)
    except RuntimeError:
        ...

    execution = flow.execution(execution_id)

    assert execution.layer(A).state.finished is True
    assert execution.layer(B).state.finished is False
    assert execution.layer(C).state.finished is False

    # Retry failed execution
    B.fail = False
    execution.resume()

    assert execution.layer(A).state.finished is True
    assert execution.layer(B).state.finished is True
    assert execution.layer(C).state.finished is True



"""Unit tests for laminar.configurations.hooks"""

from typing import Generator

from laminar.configurations import hooks


class TestAnnotation:
    def test_annotate(self) -> None:
        def func() -> Generator[None, None, None]:
            ...

        func = hooks.annotation.annotate(func, hooks.annotation.execution)
        assert hooks.annotation.get(func) == hooks.annotation.execution

    def test_execution(self) -> None:
        def func() -> Generator[None, None, None]:
            ...

        func = hooks.execution(func)
        assert hooks.annotation.get(func) == hooks.annotation.execution

    def test_retry(self) -> None:
        def func() -> Generator[None, None, None]:
            ...

        func = hooks.retry(func)
        assert hooks.annotation.get(func) == hooks.annotation.retry

    def test_submission(self) -> None:
        def func() -> Generator[None, None, None]:
            ...

        func = hooks.submission(func)
        assert hooks.annotation.get(func) == hooks.annotation.submission



"""Test parameterized flows."""

import pytest

from laminar import Flow, Layer
from laminar.components import Parameters
from laminar.configurations import datastores, executors


class ParameterFlow(Flow):
    ...


@ParameterFlow.register()
class A(Layer):
    foo: str

    def __call__(self, parameters: Parameters) -> None:
        self.foo = parameters.foo


@pytest.mark.flow
def test_flow() -> None:
    flow = ParameterFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow(foo="bar")

    assert execution.layer(Parameters).foo == "bar"
    assert execution.layer(A).foo == "bar"



"""Test branching flows."""

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors


class BranchFlow(Flow):
    ...


@BranchFlow.register()
class A(Layer):
    def __call__(self) -> None:
        self.foo = "bar"


@BranchFlow.register()
class B(Layer):
    def __call__(self, a: A) -> None:
        self.foo = a.foo


@BranchFlow.register()
class C(Layer):
    def __call__(self, a: A) -> None:
        self.foo = "baz"


@BranchFlow.register()
class D(Layer):
    def __call__(self, b: B, c: C) -> None:
        self.foo = [b.foo, c.foo]


@pytest.mark.flow
def test_flow() -> None:
    flow = BranchFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow()

    assert execution.layer(A).foo == "bar"
    assert execution.layer(B).foo == "bar"
    assert execution.layer(C).foo == "baz"
    assert execution.layer(D).foo == ["bar", "baz"]



"""Test linear flows."""

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors


class LinearFlow(Flow):
    ...


@LinearFlow.register()
class A(Layer):
    def __call__(self) -> None:
        self.foo = "bar"


@LinearFlow.register()
class B(Layer):
    def __call__(self, a: A) -> None:
        self.foo = a.foo


@LinearFlow.register()
class C(Layer):
    def __call__(self, b: B) -> None:
        self.foo = b.foo


@pytest.mark.flow
def test_flow() -> None:
    flow = LinearFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow()

    assert execution.layer(A).foo == "bar"
    assert execution.layer(B).foo == "bar"
    assert execution.layer(C).foo == "bar"



"""Tests for laminar.configurations.layers"""

from typing import Any, Dict

import pytest

from laminar import Flow, Layer
from laminar.configurations import layers
from laminar.configurations.datastores import Archive, Artifact, Record
from laminar.configurations.layers import Catch, ForEach, Parameter


class TestCatch:
    def test_success(self) -> None:
        class A(Layer):
            def __call__(self) -> None:
                raise RuntimeError

        class B(Layer):
            def __call__(self) -> None:
                raise AssertionError

        catch = Catch(RuntimeError, AssertionError)
        A(configuration=layers.Configuration(catch=catch)).execute()
        B(configuration=layers.Configuration(catch=catch)).execute()

    def test_failure(self) -> None:
        class A(Layer):
            def __call__(self) -> None:
                raise RuntimeError

        with pytest.raises(RuntimeError):
            A(configuration=layers.Configuration()).execute()

    def test_suberror(self) -> None:
        class A(Layer):
            def __call__(self) -> None:
                raise RuntimeError

        A(configuration=layers.Configuration(catch=Catch(Exception))).execute()


class TestForEach:
    @pytest.fixture(autouse=True)
    def _flow(self, flow: Flow) -> None:
        workspace: Dict[str, Any] = flow.configuration.datastore.cache
        workspace["memory:///TestFlow/archives/test-execution/A/0/foo.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="1"), Artifact(dtype="str", hexdigest="2")]
        )
        workspace["memory:///TestFlow/archives/test-execution/B/0/bar.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="3"), Artifact(dtype="str", hexdigest="4")]
        )
        workspace["memory:///TestFlow/archives/test-execution/C/0/foo.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="a")]
        )
        workspace["memory:///TestFlow/archives/test-execution/C/1/foo.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="b")]
        )
        workspace["memory:///TestFlow/archives/test-execution/C/2/foo.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="c")]
        )
        workspace["memory:///TestFlow/archives/test-execution/C/3/foo.json"] = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="d")]
        )

        workspace["memory:///TestFlow/artifacts/1.gz"] = True
        workspace["memory:///TestFlow/artifacts/2.gz"] = False
        workspace["memory:///TestFlow/artifacts/3.gz"] = 0
        workspace["memory:///TestFlow/artifacts/4.gz"] = 1

        self.flow = flow

        @self.flow.register()
        class A(Layer):
            ...

        @self.flow.register()
        class B(Layer):
            ...

        @self.flow.register(
            foreach=ForEach(parameters=[Parameter(layer=A, attribute="foo"), Parameter(layer=B, attribute="bar")])
        )
        class C(Layer):
            ...

        @self.flow.register(foreach=ForEach(parameters=[Parameter(layer=C, attribute="foo", index=None)]))
        class D(Layer):
            ...

        self.A = A
        self.B = B
        self.C = C
        self.D = D

    def test_join(self) -> None:
        layer = self.flow.layer(self.C)
        assert layer.configuration.foreach.join(layer=layer, name="foo") == Archive(
            artifacts=[
                Artifact(dtype="str", hexdigest="a"),
                Artifact(dtype="str", hexdigest="b"),
                Artifact(dtype="str", hexdigest="c"),
                Artifact(dtype="str", hexdigest="d"),
            ]
        )

    def test_grid(self) -> None:
        layer = self.flow.layer(self.C)
        assert layer.configuration.foreach.grid(layer=layer) == [
            {self.A(): {"foo": 0}, self.B(): {"bar": 0}},
            {self.A(): {"foo": 0}, self.B(): {"bar": 1}},
            {self.A(): {"foo": 1}, self.B(): {"bar": 0}},
            {self.A(): {"foo": 1}, self.B(): {"bar": 1}},
        ]

    def test_grid_none(self) -> None:
        layer = self.flow.layer(self.D)
        assert layer.configuration.foreach.grid(layer=layer) == [
            {self.C(): {"foo": 0}},
            {self.C(): {"foo": 1}},
            {self.C(): {"foo": 2}},
            {self.C(): {"foo": 3}},
        ]

    def test_splits(self) -> None:
        layer = self.flow.layer(self.C)
        assert layer.configuration.foreach.splits(layer=layer) == 4

    def test_splits_cache_hit(self) -> None:
        self.flow.configuration.datastore.cache["memory:///TestFlow/.cache/test-execution/C/.record.json"] = Record(
            flow=Record.FlowRecord("Testflow"),
            layer=Record.LayerRecord(name="C"),
            execution=Record.ExecutionRecord(splits=6),
        )

        layer = self.flow.layer(self.C)
        assert layer.configuration.foreach.splits(layer=layer) == 6

    def test_set(self) -> None:
        layer = self.flow.layer(self.C, index=0)
        A, B = layer.configuration.foreach.set(
            layer=layer, parameters=(self.flow.layer(self.A), self.flow.layer(self.B))
        )

        assert A.foo is True
        assert B.bar == 0

        layer = self.flow.layer(self.C, index=3)
        A, B = layer.configuration.foreach.set(
            layer=layer, parameters=(self.flow.layer(self.A), self.flow.layer(self.B))
        )

        assert A.foo is False
        assert B.bar == 1



"""Configurations for laminar executors."""

import asyncio
import functools
import logging
import operator
from asyncio import Task
from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Set, Tuple

from laminar.configurations import datastores, hooks
from laminar.exceptions import SchedulerError
from laminar.types import unwrap
from laminar.utils import contexts

if TYPE_CHECKING:
    from laminar import Flow, Layer


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Scheduler:
    """Base scheduler"""

    async def schedule(self, *, layer: "Layer", attempt: int = 1) -> List["Layer"]:
        """Schedule layers for execution.

        Args:
            execution: Flow execution ID
            layer: Layer to execute
            attempt: Scheduling attempt for this layer

        Returns:
            Layer splits that were executed
        """

        try:
            splits = layer.configuration.foreach.splits(layer=layer)
            tasks: List[Task["Layer"]] = []

            # Create a task per layer split
            for index in range(splits):
                instance = layer.flow.layer(layer, index=index, splits=splits, attempt=attempt)

                with hooks.event.context(layer=instance, annotation=hooks.annotation.submission):
                    tasks.append(asyncio.create_task(instance.flow.configuration.executor.submit(layer=instance)))

            # Combine all tasks into a Future so they can be waited on together
            layers = await asyncio.gather(*tasks)

        except Exception as error:
            logger.error(
                "Encountered unexpected error: %s(%s) on attempt '%d' of '%d'.",
                type(error).__name__,
                str(error),
                attempt,
                layer.configuration.retry.attempts,
            )

            # Attempt to reschedule the layer
            if attempt < layer.configuration.retry.attempts:
                with hooks.event.context(layer=layer, annotation=hooks.annotation.retry):
                    await layer.configuration.retry.sleep(layer=layer, attempt=attempt)
                return await self.schedule(layer=layer, attempt=attempt + 1)
            raise

        # Cache the layer execution metadata
        layer.flow.configuration.datastore.write_record(
            layer=layer,
            record=datastores.Record(
                flow=datastores.Record.FlowRecord(name=layer.flow.name),
                layer=datastores.Record.LayerRecord(name=layer.name),
                execution=datastores.Record.ExecutionRecord(splits=splits),
            ),
        )

        return list(layers)

    def runnable(
        self, *, dependencies: Dict[str, Set[str]], pending: Set[str], finished: Set[str]
    ) -> Tuple[Set[str], Set[str]]:
        """Find all runnable layers.

        Args:
            dependencies: Layer dependencies
            pending: Pending layers
            finished: Finished layers

        Returns:
            * Remaining pending layers
            * Runnable layers
        """

        runnable = {layer for layer in pending if dependencies[layer].issubset(finished)}
        return pending - runnable, runnable

    def skippable(self, *, flow: "Flow", runnable: Set[str], finished: Set[str]) -> Tuple[Set[str], Set[str]]:
        """Find all skippable layers.

        Args:
            flow: Flow being scheduled.
            runnable: Runnable layers.
            finished: Finished layers.

        Returns:
            * Runnable layers
            * Finished layers
        """

        skippable: Set[str] = set()

        for layer in runnable:
            instance = flow.layer(layer)

            # Check for retries
            if instance.flow.execution.retry and instance.state.finished:
                skippable.add(layer)
                continue

            # Check entry hooks
            conditions = hooks.condition.gather(layer=instance, annotation=hooks.annotation.entry)
            if not functools.reduce(operator.and_, conditions, True):
                skippable.add(layer)
                continue

            # Check dependencies if not conditions exist
            if not conditions and not all(dependency.state.finished for dependency in instance._dependencies):
                skippable.add(layer)
                continue

        if skippable:
            logger.info("Skipping layers: %s", sorted(skippable))
        return runnable - skippable, finished | skippable

    def running(
        self, *, flow: "Flow", runnable: Set[str], running: Set["Task[List[Layer]]"]
    ) -> Set["Task[List[Layer]]"]:
        """Schedule runnable layers.

        Args:

            flow: Flow that the layers are being run in.
            runnalbe: Runnable layers.
            running: Currently running layers.

        Returns:
            Async tasks for new and existing running layers.
        """

        return running | {asyncio.create_task(self.schedule(layer=flow.layer(layer))) for layer in runnable}

    async def wait(
        self, *, running: Set["Task[List[Layer]]"], finished: Set[str], condition: str
    ) -> Tuple[Set["Task[List[Layer]]"], Set[str]]:
        """Wait on the completion of running layers.

        Args:
            running: Running layers.
            finished: Finished layers
            condition: Condition to wait on.

        Returns:
            * Remaining running layers
            * Finished layers
        """

        # Wait until the first task completes
        completed, incomplete = await asyncio.wait(running, return_when=condition)

        # Add all completed tasks to finished tasks
        running = set(incomplete)
        finished = finished | {(await task)[0].name for task in completed}

        return running, finished

    @contexts.EventLoop
    async def loop(self, *, flow: "Flow", dependencies: Dict[str, Set[str]], finished: Set[str]) -> None:
        """Run the scheduling loop.

        Args:
            flow: Flow to schedule.
            dependencies: Dependencies in the flow to schedule.
            finished: Finished layers that don't need to be scheduled.

        Raises:
            SchedulerError: If no layers are runnable and there are no running or pending layers.
        """

        dependencies = deepcopy(dependencies)
        finished = deepcopy(finished)

        logger.info("Flow: '%s'", flow.name)
        logger.info("Execution: '%s'", unwrap(flow.execution.id))
        logger.info("Dependencies: '%s'", dependencies)

        pending = set(dependencies) - finished
        runnable: Set[str] = set()
        running: Set[Task[List["Layer"]]] = set()

        # Start the scheduling loop
        while pending:
            logger.info("Pending layers: %s", sorted(pending))
            pending, runnable = self.runnable(dependencies=dependencies, pending=pending, finished=finished)

            # There are pending layers but nothing is runnable or running.
            if not runnable and not running and pending:
                raise SchedulerError(
                    f"Stuck waiting to schedule: {sorted(pending)}."
                    f" Finished layers: {sorted(finished)}."
                    f" Remaining dependencies: { {task: sorted(dependencies[task]) for task in sorted(pending)} }"
                )

            logger.info("Runnable layers: %s", sorted(runnable))
            runnable, finished = self.skippable(flow=flow, runnable=runnable, finished=finished)

            running = self.running(flow=flow, runnable=runnable, running=running)

            # Determine async task wait condition
            condition = asyncio.FIRST_COMPLETED if pending else asyncio.ALL_COMPLETED

            logger.info("Running layers: %s", sorted(set(dependencies) - pending - finished))
            if running:
                running, finished = await self.wait(running=running, finished=finished, condition=condition)
            logger.info("Finished layers: %s", sorted(finished))

    def compile(self, *, flow: "Flow") -> Dict[str, Any]:
        """Compile an intermediate representation of the Flow."""

        raise NotImplementedError

    def create(self, *, ir: Dict[str, Any]) -> None:
        """Create a delegated scheduler to schedule the Flow."""

        raise NotImplementedError

    def invoke(self) -> None:
        """Invoke the delegated scheduler to start a Flow execution."""

        raise NotImplementedError


@dataclass(frozen=True)
class Delegated(Scheduler):
    """Scheduler that compiles, creates, and invokes a Flow execution on another scheduler."""


class AWS:
    @dataclass(frozen=True)
    class StepFunctions(Scheduler):
        """Schedule flows in AWS Step Functions.

        Usage::

            Flow(scheduler=AWS.StepFunctions())
        """



"""Unit tests for laminar.configurations.schedulers"""

from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator
from unittest.mock import Mock, patch

import pytest

from laminar import Layer
from laminar.configurations.datastores import Record
from laminar.configurations.schedulers import Scheduler


@asynccontextmanager
async def coroutine(path: str) -> AsyncGenerator[Mock, None]:
    async def func(*args: Any, **kwargs: Any) -> None:
        ...

    with patch(path) as mock:
        mock.return_value = func
        yield mock


@pytest.mark.asyncio
class TestScheduler:
    scheduler = Scheduler()

    async def test_schedule(self, layer: Layer) -> None:
        async with coroutine("laminar.configurations.executors.Thread.submit") as mock_execute:
            await self.scheduler.schedule(layer=layer)

            mock_execute.assert_called_once_with(layer=layer)

        assert layer.flow.configuration.datastore.read_record(layer=layer) == Record(
            flow=Record.FlowRecord(name="TestFlow"),
            layer=Record.LayerRecord(name="Layer"),
            execution=Record.ExecutionRecord(splits=1),
        )



import asyncio
import contextlib
import os
from typing import Any, Dict, TypeVar, cast


class Attributes(contextlib.ContextDecorator):
    def __init__(self, cls: object, **attributes: Any) -> None:
        """Modify a class's attributes.

        Args:
            cls (object): Class to modify
            **attributes (Any): Attribute key/value pairs to modify.
        """

        self.cls = cls
        self.attributes = attributes
        self._attributes: Dict[str, Any] = {}

    def __enter__(self) -> "Attributes":
        """Modify the class attributes."""

        for key, value in self.attributes.items():
            if hasattr(self.cls, key):
                self._attributes[key] = getattr(self.cls, key, None)
            setattr(self.cls, key, value)

        return self

    def __exit__(self, *_: Any) -> None:
        """Revert the class attribute changes."""

        for key in self.attributes:
            if key in self._attributes:
                setattr(self.cls, key, self._attributes[key])
            else:
                delattr(self.cls, key)


class Environment(contextlib.ContextDecorator):
    def __init__(self, **variables: Any) -> None:
        self.variables = {key: str(value) for key, value in variables.items()}

    def __enter__(self) -> "Environment":
        """Modify environment"""

        self.old = dict(os.environ)
        os.environ.update(self.variables)
        return self

    def __exit__(self, *_: Any) -> None:
        """Revert environment changes"""

        os.environ.clear()
        os.environ.update(self.old)


T = TypeVar("T", bound=Any)


def EventLoop(func: T) -> T:
    """Execute the decorated function in an asyncio event loop.

    Usage::

        from laminar.utils import contexts

        @contexts.EventLoop
        async def main() -> None:
            ...
    """

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        asyncio.run(func(*args, **kwargs))

    return cast(T, wrapper)



"""Core components for build flows."""

import copy
import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import Any, Callable, Dict, Iterable, Optional, Set, Tuple, Type, TypeVar, Union, overload

from ksuid import KsuidMs

from laminar.configurations import datastores, executors, flows, hooks, layers, schedulers
from laminar.exceptions import ExecutionError, FlowError
from laminar.settings import current
from laminar.types import LayerType, hints, unwrap
from laminar.utils import contexts, stringify

logger = logging.getLogger(__name__)

FLOW_RESERVED_KEYWORDS = {"configuration", "execution"}
LAYER_RESERVED_KEYWORDS = {"attempt", "configuration", "flow", "index", "namespace", "splits", "state"}


@dataclass
class Layer:
    """Task to execute as part of a flow.

    Usage::

        from laminar import Layer

        class Task(Layer):
            ...
    """

    #: Configurations for the Layer
    configuration: layers.Configuration
    #: Flow the Layer is registered to
    flow: "Flow"
    #: Layer state
    state: layers.State

    #: Current layer execution attempt
    attempt: Optional[int] = current.layer.attempt
    #: Layer index in its splits
    index: Optional[int] = current.layer.index
    #: Number of splits in the layer execution
    splits: Optional[int] = current.layer.splits

    def __init__(self, **attributes: Any) -> None:
        for key, value in attributes.items():
            setattr(self, key, value)
        self.state = layers.State(layer=self)

    __call__: Callable[..., None]  # type: ignore

    def __call__(self) -> None:  # type: ignore
        ...

    def __repr__(self) -> str:
        return stringify(self, self.name, "flow", "index", "splits")

    def __deepcopy__(self, memo: Dict[int, Any]) -> "Layer":
        cls = self.__class__
        result: Layer = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, Layer):
            return type(self) is type(other) and self.name == other.name
        else:
            return False

    def __getattr__(self, name: str) -> Any:
        if name not in self.__dict__:
            try:
                self.__dict__[name] = self.flow.configuration.datastore.read_artifact(
                    layer=self, archive=self.configuration.foreach.join(layer=self, name=name)
                )
            except RecursionError as error:
                raise AttributeError(f"Object '{self.name}' has no attribute '{name}'.") from error
            except FileNotFoundError as error:
                message = f"Object '{self.name}' has no attribute '{name}'."
                if not self.state.finished:
                    message = message + f" Layer {self.name} was not finished."
                raise AttributeError(message) from error

        return self.__dict__[name]

    def __getstate__(self) -> Dict[str, Any]:
        return self.__dict__

    def __hash__(self) -> int:
        return hash(self.name)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.name < other
        elif isinstance(other, Layer):
            return self.name < other.name
        else:
            raise TypeError(f"Type '{type(self).__name__}' and type '{type(other).__name__}' are incomparable.")

    def __setstate__(self, slots: Dict[str, Any]) -> None:
        self.__dict__ = slots

    @property
    def artifacts(self) -> Dict[str, Any]:
        """Artifacts assigned to the layer."""

        return {artifact: value for artifact, value in vars(self).items() if artifact not in LAYER_RESERVED_KEYWORDS}

    @property
    def _hooks(self) -> Set[Callable[..., Any]]:
        return {
            hook
            for hook in chain(vars(type(self.flow)).values(), vars(type(self)).values())
            if hooks.annotation.get(hook) is not None
        }

    @property
    def _parameters(self) -> Dict[str, Tuple["Layer", ...]]:
        return {
            self.__call__.__name__: hints(self.flow, self.__call__),
            **{hook.__name__: hints(self.flow, hook) for hook in self._hooks},
        }

    @property
    def _dependencies(self) -> Set["Layer"]:
        return set(chain.from_iterable(self._parameters.values()))

    @property
    def dependencies(self) -> Set[str]:
        """Layers this layer depends on."""

        return {layer.name for layer in self._dependencies}

    @property
    def hooks(self) -> Dict[str, Set[Callable[..., Any]]]:
        """Hooks attached to this layer collated by annotation."""

        _hooks: Dict[str, Set[Callable[..., Any]]] = defaultdict(set)
        for hook in self._hooks:
            _hooks[unwrap(hooks.annotation.get(hook))].add(hook)
        return _hooks

    @property
    def name(self) -> str:
        """Name of the Layer"""

        return type(self).__name__

    def execute(self, *parameters: "Layer") -> None:
        """Execute a layer.

        Args:
            *parameters: Input layers to the layer.
        """

        # Attempt to write any existing layer artifacts before failing
        try:
            with self.configuration.catch:
                self(*parameters)
        finally:
            for artifact, value in self.artifacts.items():
                self.flow.configuration.datastore.write(layer=self, name=artifact, values=[value])

    def shard(self, **artifacts: Iterable[Any]) -> None:
        """Store each item of a sequence separately so that they may be loaded individually downstream.

        Notes:
            The artifact that is loaded is of type laminar.configurations.datastores.Accessor.

        Usage::

            class Task(Layer):
                def __call__(self) -> None:
                    self.shard(foo=["a", "b", "c"])

        Args:
            **artifacts: Sequence to break up and store.
        """

        for artifact, sequence in artifacts.items():
            self.flow.configuration.datastore.write(layer=self, name=artifact, values=sequence)


class Parameters(Layer):
    """Special Layer for handling Flow parameters."""


T = TypeVar("T", bound=Layer)


class Flow:
    """Collection of tasks that execute in a specific order.

    Usage::

        from laminar import Flow, Layer

        class HelloFlow(Flow):
            ...
    """

    registry: Dict[str, Layer]
    """Layers registered with the flow"""

    def __init__(
        self,
        *,
        datastore: datastores.DataStore = datastores.Local(),
        executor: executors.Executor = executors.Docker(),
        scheduler: schedulers.Scheduler = schedulers.Scheduler(),
    ) -> None:
        """
        Args:
            datastore: Datastore to execute the flow with. Optional; Defaults to datastores.Local().
            executor: Executor to run layers with. Optional; Defaults to executors.Docker().
            scheduler: Scheduler to manage the flow with. Optional; Defaults to schedulers.Scheduler().

        Raises:
            FlowError: If the flow is used to configure the Memory datastore without a Thread executor.
        """

        self.name = type(self).__name__
        self.execution = flows.Execution(id=current.execution.id, flow=self)

        if isinstance(datastore, datastores.Memory) and not isinstance(executor, executors.Thread):
            raise FlowError("The Memory datastore can only be used with the Thread executor.")

        self.configuration = flows.Configuration(datastore=datastore, executor=executor, scheduler=scheduler)

    def __init_subclass__(cls) -> None:
        flow: Type[Flow]
        layer: Layer

        # Register all subflow layers with this layer
        cls.registry = {Parameters().name: Parameters(configuration=layers.Configuration())}
        for flow in cls.__bases__:
            for name, layer in getattr(flow, "registry", {}).items():
                if name != Parameters().name:
                    callback = cls.register(
                        container=layer.configuration.container,
                        foreach=layer.configuration.foreach,
                        retry=layer.configuration.retry,
                    )
                    callback(layer.__class__)

    @property
    def _dependencies(self) -> Dict[Layer, Set[Layer]]:
        return {
            self.layer(child): {self.layer(parent) for parent in parents}
            for child, parents in self.dependencies.items()
        }

    @property
    def dependencies(self) -> Dict[str, Set[str]]:
        """A mapping of each layer and the layers it depends on."""

        return {layer: self.layer(layer).dependencies for layer in self.registry}

    @property
    def _dependents(self) -> Dict[Layer, Set[Layer]]:
        return {
            self.layer(parent): {self.layer(child) for child in children}
            for parent, children in self.dependents.items()
        }

    @property
    def dependents(self) -> Dict[str, Set[str]]:
        """A mapping of each layer and the layers that depend on it."""

        dependents: Dict[str, Set[str]] = defaultdict(set)
        for child, parents in self.dependencies.items():
            for parent in parents:
                dependents[parent].add(child)
        return dependents

    def __call__(self, *, execution: Optional[str] = None, **attributes: Any) -> flows.Execution:
        """Execute the flow or execute a layer in the flow.

        Notes:
            If the execution id and layer name are set, execute a layer.
            Else execute the flow.

        Usage::

            class HelloFlow(Flow):
                ...
            flow = HelloFlow()

            flow()
            flow("execution-id")
        """

        # Execute a layer in the flow.
        if self.execution.id is not None and self.name == current.flow.name and current.layer.name in self.registry:
            execution = self.execution.id
            self.execute(execution=self.execution.id, layer=self.layer(current.layer.name))

        # Schedule execution of the flow.
        elif self.execution.id is None:
            self.execution = flows.Execution(id=execution or str(KsuidMs()), flow=self)
            self.parameters(execution=self.execution.id, **attributes)
            self.schedule(execution=unwrap(self.execution.id), dependencies=self.dependencies)

        return self.execution

    def __repr__(self) -> str:
        return stringify(self, self.name, "execution")

    def execute(self, *, execution: str, layer: Layer) -> None:
        """Execute a single layer of the flow.

        Usage::

            class ExecutionFlow(Flow):
                ...

            @ExecutionFlow.register()
            class A(Layer):
                ...

            flow = ExecutionFlow()
            flow.execute(execution="test-execution", layer=flow.layer(A, index=0, splits=2))

        Args:
            execution: ID of the execution being run.
            layer: Layer of the flow to execute.
        """

        with contexts.Attributes(layer.flow.execution, id=execution):
            logger.info("Starting layer '%s'.", layer.name if layer.splits == 1 else f"{layer.name}/{layer.index}")

            # Setup the Layer parameter values
            parameters = layer.configuration.foreach.set(layer=layer, parameters=layer._parameters["__call__"])

            with hooks.event.context(layer=layer, annotation=hooks.annotation.execution):
                layer.execute(*parameters)

            logger.info("Finishing layer '%s'.", layer.name if layer.splits == 1 else f"{layer.name}/{layer.index}")

    def schedule(self, *, execution: str, dependencies: Dict[str, Set[str]]) -> None:
        """Schedule layers to run in sequence in the flow.

        Args:
            execution: ID of the execution being run.
            dependencies: Mapping of layers to layers it depends on.
        """

        with contexts.Attributes(self.execution, id=execution):
            self.configuration.scheduler.loop(  # type: ignore
                flow=self, dependencies=dependencies, finished={Parameters().name}
            )

    @classmethod
    def register(
        cls,
        catch: layers.Catch = layers.Catch(),
        container: layers.Container = layers.Container(),
        foreach: layers.ForEach = layers.ForEach(),
        retry: layers.Retry = layers.Retry(),
    ) -> Callable[[LayerType], LayerType]:
        """Add a layer to the flow.

        Usage::

            @Flow.register()
            class Task(Layer):
                ...
        """

        def wrapper(Layer: LayerType) -> LayerType:
            layer = Layer(
                configuration=layers.Configuration(catch=catch, container=container, foreach=foreach, retry=retry)
            )

            if layer.name in cls.registry:
                raise FlowError(
                    f"Duplicate layer added to flow '{cls.__name__}'.\n"
                    f"  Given layer '{layer.name}'.\n"
                    f"  Added layers {sorted(cls.registry)}"
                )

            # First register the layer without the flow attribute
            cls.registry[layer.name] = copy.deepcopy(layer)

            return Layer

        return wrapper

    @overload
    def layer(self, layer: str, **atributes: Any) -> Layer:
        ...

    @overload
    def layer(self, layer: Type[T], **attributes: Any) -> T:
        ...

    @overload
    def layer(self, layer: T, **attributes: Any) -> T:
        ...

    def layer(self, layer: Union[str, Type[Layer], Layer], **attributes: Any) -> Layer:
        """Get a registered flow layer.

        Usage::

            flow.layer("A")
            flow.layer(A)
            flow.layer(A())
            flow.layer(A(), index=0, splits=2)

        Args:
            layer: Layer to get.
            **attributes: Keyword attributes to add to the Layer.

        Returns:
            Layer that is registered to the flow.
        """

        if isinstance(layer, Layer):
            layer = layer.name
        elif not isinstance(layer, str):
            layer = layer().name

        # Deepcopy so that layer artifacts don't mess with other layer split executions
        layer = copy.deepcopy(self.registry[layer])

        # Inject the flow attribute to link the layer to the flow
        layer.flow = self
        for key, value in attributes.items():
            setattr(layer, key, value)

        return layer

    def parameters(self, *, execution: Optional[str] = None, **artifacts: Any) -> str:
        """Configure parameters for a flow execution

        Usage::

            execution = flow.parameters(foo="bar")
            flow(execution=execution)

        Args:
            execution: ID of the execution to configure the flow parameters for.
            artifacts: Key/value pairs of parameters to add to the flow.

        Returns:
            ID of the execution the parameters were added to.
        """

        execution = execution or str(KsuidMs())
        with contexts.Attributes(self.execution, id=execution):
            # Property setup the layer for writing to the datastore
            layer = self.layer(Parameters, index=0, splits=1, attempt=0)

            # Assign artifact values
            for name, value in artifacts.items():
                if name in LAYER_RESERVED_KEYWORDS:
                    raise ExecutionError(
                        "A flow parameter is in the list of layer reserved keywords:"
                        f" {sorted(LAYER_RESERVED_KEYWORDS)}. Given {name}."
                    )
                setattr(layer, name, value)

            # Fake an execution to write the artifacts to the datastore.
            self.execute(execution=execution, layer=layer)

            # Record parameter layer execution
            layer.flow.configuration.datastore.write_record(
                layer=layer,
                record=datastores.Record(
                    flow=datastores.Record.FlowRecord(name=layer.flow.name),
                    layer=datastores.Record.LayerRecord(name=layer.name),
                    execution=datastores.Record.ExecutionRecord(splits=unwrap(layer.splits)),
                ),
            )

        return execution



"""Test conditional branching."""

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors, hooks


class ConditionalFlow(Flow):
    ...


@ConditionalFlow.register()
class A(Layer):
    def __call__(self) -> None:
        self.foo = "bar"


# Always skip B
@ConditionalFlow.register()
class B(Layer):
    def __call__(self, a: A) -> None:
        self.foo = a.foo

    @hooks.entry
    def failure(self) -> bool:
        return False


# C executes because A executed
@ConditionalFlow.register()
class C(Layer):
    def __call__(self, a: A) -> None:
        self.foo = "baz"


# D skips because B was skipped
@ConditionalFlow.register()
class D(Layer):
    def __call__(self, b: B) -> None:
        self.foo = b.foo


# Force E to execute even though D skipped
@ConditionalFlow.register()
class E(Layer):
    def __call__(self, c: C, d: D) -> None:
        if d.state.finished:
            self.foo = [c.foo, d.foo]
        else:
            self.foo = [c.foo]

    @hooks.entry
    def success(self) -> bool:
        return True


@pytest.mark.flow
def test_flow() -> None:
    flow = ConditionalFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow()

    assert execution.layer(A).state.finished is True
    assert execution.layer(A).foo == "bar"
    assert execution.layer(B).state.finished is False
    assert execution.layer(C).state.finished is True
    assert execution.layer(C).foo == "baz"
    assert execution.layer(D).state.finished is False
    assert execution.layer(E).state.finished is True
    assert execution.layer(E).foo == ["baz"]



"""Test foreach flows."""

from typing import List, cast

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors
from laminar.configurations.layers import ForEach, Parameter
from laminar.types import unwrap


class ForeachFlow(Flow):
    ...


@ForeachFlow.register()
class A(Layer):
    foo: List[int]

    def __call__(self) -> None:
        self.shard(foo=[1, 2, 3])


@ForeachFlow.register(foreach=ForEach(parameters=[Parameter(layer=A, attribute="foo")]))
class B(Layer):
    foo: List[int]

    def __call__(self, a: A) -> None:
        self.foo = cast(List[int], cast(int, a.foo) + unwrap(self.index) ** 2)


@ForeachFlow.register()
class C(Layer):
    def __call__(self, b: B) -> None:
        self.foo = [value + i for i, value in enumerate(b.foo)]


@pytest.mark.flow
def test_flow() -> None:
    flow = ForeachFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow()

    assert list(execution.layer(A).foo) == [1, 2, 3]
    assert list(execution.layer(B).foo) == [1, 3, 7]
    assert execution.layer(C).foo == [1, 4, 9]



"""Tests for laminar.configurations.datastores"""

import io
import json
from pathlib import Path
from typing import Any, Dict, List, cast
from unittest.mock import Mock, call, mock_open, patch

import cloudpickle
import pytest

from laminar import Layer
from laminar.configurations.datastores import Accessor, Archive, Artifact, DataStore, Local, Record


class TestArtifact:
    artifact = Artifact(dtype="str", hexdigest="foo")

    def test_dict(self) -> None:
        assert self.artifact.dict() == {"dtype": "str", "hexdigest": "foo"}

    def test_path(self, layer: Layer) -> None:
        assert self.artifact.path(layer=layer) == f"{layer.flow.name}/artifacts/foo.gz"


class TestArchive:
    archive = Archive(artifacts=[Artifact(dtype="str", hexdigest="foo"), Artifact(dtype="str", hexdigest="bar")])

    def test_dict(self) -> None:
        assert self.archive.dict() == {
            "artifacts": [{"dtype": "str", "hexdigest": "foo"}, {"dtype": "str", "hexdigest": "bar"}]
        }

    def test_len(self) -> None:
        assert len(self.archive) == 2

    def test_path(self, layer: Layer) -> None:
        assert (
            self.archive.path(layer=layer, index=0, name="test-archive", cache=False)
            == f"{layer.flow.name}/archives/{layer.flow.execution.id}/{layer.name}/{layer.index}/test-archive.json"
        )
        assert (
            self.archive.path(layer=layer, index=0, name="test-archive", cache=True)
            == f"{layer.flow.name}/.cache/{layer.flow.execution.id}/{layer.name}/test-archive.json"
        )

    def test_parse(self) -> None:
        expected = {"artifacts": [{"dtype": "str", "hexdigest": "foo"}, {"dtype": "str", "hexdigest": "bar"}]}
        assert Archive.parse(expected).dict() == expected


class TestAccessor:
    @pytest.fixture(autouse=True)
    def _accessor(self, layer: Layer) -> None:
        workspace: Dict[str, Any] = layer.flow.configuration.datastore.cache
        workspace["memory:///TestFlow/artifacts/test-hexdigest-0.gz"] = "foo"
        workspace["memory:///TestFlow/artifacts/test-hexdigest-1.gz"] = "bar"

        self.accessor = Accessor(
            archive=Archive(
                artifacts=[
                    Artifact(dtype="str", hexdigest="test-hexdigest-0"),
                    Artifact(dtype="str", hexdigest="test-hexdigest-1"),
                ]
            ),
            layer=layer,
        )

    def test_len(self) -> None:
        assert len(self.accessor) == 2

    def test_get_item(self) -> None:
        assert [self.accessor[0], self.accessor[1]] == ["foo", "bar"]

    def test_iterable(self) -> None:
        assert list(self.accessor) == ["foo", "bar"]

    def test_for(self) -> None:
        items: List[str] = []
        for item in self.accessor:
            items.append(item)
        assert items == ["foo", "bar"]

    def test_slice(self) -> None:
        assert self.accessor[:1] == ["foo"]

    def test_index_out_of_bounds(self) -> None:
        with pytest.raises(IndexError):
            self.accessor[10]

    def test_index_other(self) -> None:
        with pytest.raises(TypeError):
            self.accessor[cast(int, "a")]


class TestDatastore:
    @pytest.fixture(autouse=True)
    def _archive(self) -> None:
        self.archive = Archive(
            artifacts=[Artifact(dtype="str", hexdigest="foo"), Artifact(dtype="str", hexdigest="bar")]
        )

    @pytest.fixture(autouse=True)
    def _datastore(self) -> None:
        self.datastore = DataStore(root="path/to/root/")

    @pytest.fixture(autouse=True)
    def _record(self) -> None:
        self.record = Record(
            flow=Record.FlowRecord(name="test-flow"),
            layer=Record.LayerRecord(name="test-layer"),
            execution=Record.ExecutionRecord(splits=2),
        )

    def test_init(self) -> None:
        assert self.datastore.root == "path/to/root"

    def test_uri(self) -> None:
        assert self.datastore.uri(path="other/path") == "path/to/root/other/path"

    @patch("laminar.utils.fs.exists")
    def test_exists(self, mock_exists: Mock) -> None:
        self.datastore.exists(path="path/to/file")

        mock_exists.assert_called_once_with(uri="path/to/root/path/to/file")

    def test_protocol(self) -> None:
        mock_protocol = Mock()
        self.datastore.protocol(str)(mock_protocol)
        assert self.datastore.protocols["builtins.str"] == mock_protocol.return_value

    @patch("laminar.utils.fs.open")
    def test_read_archive(self, mock_open: Mock, layer: Layer) -> None:
        mock_open.return_value = io.StringIO(json.dumps(self.archive.dict()))

        assert self.datastore.read_archive(layer=layer, index=0, name="test-archive") == self.archive

        mock_open.assert_called_once_with(
            "path/to/root/TestFlow/archives/test-execution/Layer/0/test-archive.json", "rb"
        )

    @patch("laminar.utils.fs.open")
    def test_read_artifact(self, mock_open: Mock, layer: Layer) -> None:
        mock_open.return_value = io.BytesIO(cloudpickle.dumps("test-value"))

        assert (
            self.datastore.read_artifact(
                layer=layer, archive=Archive(artifacts=[Artifact(dtype="str", hexdigest="foo")])
            )
            == "test-value"
        )

        mock_open.assert_called_once_with("path/to/root/TestFlow/artifacts/foo.gz", "rb")

    def test_read_artifact_accessor(self, layer: Layer) -> None:
        assert self.datastore.read_artifact(layer=layer, archive=self.archive) == Accessor(
            archive=self.archive, layer=layer
        )

    @patch("laminar.configurations.datastores.DataStore._read")
    def test_read(self, mock_read: Mock, layer: Layer) -> None:
        mock_read.return_value.__len__.return_value = 1

        self.datastore.read(layer=layer, index=0, name="test")

        assert mock_read.call_args_list == [
            call(
                uri="path/to/root/TestFlow/archives/test-execution/Layer/0/test.json",
                dtype="laminar.configurations.datastores.ArchiveProtocol",
            ),
            call(
                uri=self.datastore.uri(path=mock_read.return_value.artifacts[0].path.return_value),
                dtype=mock_read.return_value.artifacts[0].dtype,
            ),
        ]

    @patch("laminar.utils.fs.open", new_callable=mock_open)
    def test_write_archive(self, mock_open: Mock, layer: Layer) -> None:
        assert (
            self.datastore.write_archive(layer=layer, name="test-archive", artifacts=self.archive.artifacts)
            == self.archive
        )

        mock_open.assert_called_once_with(
            "path/to/root/TestFlow/archives/test-execution/Layer/0/test-archive.json", "wb"
        )
        mock_open.return_value.write.assert_called_once_with(
            b'{"artifacts": [{"dtype": "str", "hexdigest": "foo"}, {"dtype": "str", "hexdigest": "bar"}]}'
        )

    @patch("laminar.utils.fs.open", new_callable=mock_open)
    def test_write_artifact(self, mock_open: Mock, layer: Layer) -> None:
        assert self.datastore.write_artifact(layer=layer, value="test-value") == Artifact(
            dtype="builtins.str", hexdigest="7d3d5dd741934c11ce55c08d83052780db2f29438238f602afbd51b177a98b7f"
        )

        mock_open.assert_called_once_with(
            "path/to/root/TestFlow/artifacts/7d3d5dd741934c11ce55c08d83052780db2f29438238f602afbd51b177a98b7f.gz", "wb"
        )
        mock_open.return_value.write.assert_called_once_with(
            b"\x80\x05\x95\x0e\x00\x00\x00\x00\x00\x00\x00\x8c\ntest-value\x94."
        )

    @patch("laminar.configurations.datastores.DataStore._write")
    def test_write(self, mock_write: Mock, layer: Layer) -> None:
        self.datastore.write(layer=layer, name="test-artifact", values=[True])

        assert mock_write.call_args_list == [
            call(
                value=True,
                uri=(
                    "path/to/root/TestFlow/artifacts/"
                    "5280fce43ea9afbd61ec2c2a16c35118af29eafa08aa2f5f714e54dc9cceb5ae.gz"
                ),
                dtype="builtins.bool",
            ),
            call(
                value=Archive(
                    artifacts=[
                        Artifact(
                            dtype="builtins.bool",
                            hexdigest="5280fce43ea9afbd61ec2c2a16c35118af29eafa08aa2f5f714e54dc9cceb5ae",
                        )
                    ]
                ),
                uri="path/to/root/TestFlow/archives/test-execution/Layer/0/test-artifact.json",
                dtype="laminar.configurations.datastores.ArchiveProtocol",
            ),
        ]

    @patch("laminar.utils.fs.open")
    def test_read_record(self, mock_open: Mock, layer: Layer) -> None:
        mock_open.return_value = io.StringIO(json.dumps(self.record.dict()))

        assert self.datastore.read_record(layer=layer) == self.record

        mock_open.assert_called_once_with("path/to/root/TestFlow/.cache/test-execution/Layer/.record.json", "rb")

    @patch("laminar.utils.fs.open", new_callable=mock_open)
    def test_write_record(self, mock_write: Mock, layer: Layer) -> None:
        self.datastore.write_record(layer=layer, record=self.record)

        mock_write.assert_called_once_with("path/to/root/TestFlow/.cache/test-execution/Layer/.record.json", "wb")
        mock_write.return_value.write.assert_called_once_with(
            b'{"flow": {"name": "test-flow"}, "layer": {"name": "test-layer"}, "execution": {"splits": 2}}'
        )


class TestLocal:
    @pytest.fixture(autouse=True)
    def _datastore(self, tmp_path: Path) -> None:
        self.datastore = Local(root=str(tmp_path))

    def test_read_write(self, layer: Layer) -> None:
        self.datastore.write(layer=layer, name="test", values=[[True, False]])
        assert self.datastore.read(layer=layer, index=0, name="test") == [True, False]



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

README = "README.md"
REQUIREMENTS = "requirements.txt"
VERSION = "laminar/version.py"

with open(README, "r") as readme, open(REQUIREMENTS, "r") as requirements:
    setup(
        author="Ryan Chui",
        author_email="ryan.w.chui@gmail.com",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "Intended Audience :: System Administrators",
            "License :: OSI Approved :: MIT License",
            "Operating System :: Unix",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Internet",
            "Typing :: Typed",
        ],
        description="Modern, container first framework for creating ready for production workflows.",
        include_package_data=True,
        install_requires=requirements.read(),
        license="MIT",
        long_description=readme.read(),
        long_description_content_type="text/markdown",
        name="laminar",
        packages=find_packages(exclude=["tests"]),
        project_urls={
            "Homepage": "https://github.com/rchui/laminar",
            "Documentation": "https://rchui.github.io/laminar/html/index.html",
        },
        python_requires=">=3.8",
        url="https://github.com/rchui/laminar",
        version=str(SourceFileLoader("version", VERSION).load_module().VERSION),
    )



"""Configurations for laminar data sources."""

import hashlib
import json
import re
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, Callable, Dict, Generator, Iterable, List, Tuple, Type, Union, overload

import boto3
from dacite.core import from_dict

from laminar.configurations import serde
from laminar.types import unwrap
from laminar.utils import fs

if TYPE_CHECKING:
    from laminar import Flow, Layer
    from laminar.configurations.flows import Execution
else:
    Execution = "Execution"
    Flow = "Flow"
    Layer = "Layer"

DEFAULT_SERDE = serde.PickleProtocol()

ARCHIVE_PATTERN = re.compile(
    r"^.+"  # Greedily match from start
    r"\/(?P<flow>.+?)"  # Match flow name
    r"\/archives"  # Match archive directory
    r"\/(?P<execution>.+?)"  # Match execution id
    r"(?:\/(?P<layer>.+?))?"  # Match layer name
    r"(?:\/(?P<split>\d+?))?"  # Match split index
    r"(?:\/(?P<artifact>.+?)\.json)?$"  # Match artifact name
)


@dataclass(frozen=True)
class Record:
    """Handler for metadata about how a Layer was executed."""

    @dataclass(frozen=True)
    class FlowRecord:
        #: Name of the flow
        name: str

    @dataclass(frozen=True)
    class LayerRecord:
        #: Name of the layer
        name: str

    @dataclass(frozen=True)
    class ExecutionRecord:
        #: Number of splits in the layer
        splits: int

    #: Flow record information
    flow: FlowRecord
    #: Layer record information
    layer: LayerRecord
    #: Execution record information
    execution: ExecutionRecord

    @staticmethod
    def path(*, layer: Layer) -> str:
        """Get the path to the Record."""

        return fs.join(layer.flow.name, ".cache", unwrap(layer.flow.execution.id), layer.name, ".record.json")

    def dict(self) -> Dict[str, Any]:
        """Convert the Record to a dict."""

        return asdict(self)

    @staticmethod
    def parse(source: Dict[str, Any]) -> "Record":
        """Get a Record from a dict."""

        return from_dict(Record, source)


class RecordProtocol(serde.Protocol):
    """Custom protocol for serializing Records."""

    def load(self, file: BinaryIO) -> Record:
        return Record.parse(json.load(file))

    def dumps(self, value: Record) -> bytes:
        return json.dumps(value.dict()).encode()


@dataclass(frozen=True)
class Artifact:
    """Handler for artifacts in the laminar datastore.

    Notes:
        Artifacts are gziped, pickled layer instance attributes.
    """

    #: Data type of the artifact.
    dtype: str
    #: SHA256 hexdigest of the artifact bytes.
    hexdigest: str

    def path(self, *, layer: Layer) -> str:
        """Get the path to the Artifact."""

        return fs.join(layer.flow.name, "artifacts", f"{self.hexdigest}.gz")

    def dict(self) -> Dict[str, str]:
        """Convert the Artifact to a dict."""

        return asdict(self)


@dataclass(frozen=True)
class Archive:
    """Handler for archives in the laminar datastore.

    Notes:
        Archives are metadata for artifacts.
    """

    artifacts: List[Artifact]

    def __len__(self) -> int:
        return len(self.artifacts)

    @staticmethod
    def path(*, layer: Layer, index: int, name: str, cache: bool = False) -> str:
        """Get the path to the Archive."""

        parts: Tuple[str, ...]

        if cache:
            parts = (layer.flow.name, ".cache", unwrap(layer.flow.execution.id), layer.name, f"{name}.json")
        else:
            parts = (
                layer.flow.name,
                "archives",
                unwrap(layer.flow.execution.id),
                layer.name,
                str(index),
                f"{name}.json",
            )

        return fs.join(*parts)

    def dict(self) -> Dict[str, List[Dict[str, str]]]:
        """Convert the Archive to a dict."""

        return asdict(self)

    @staticmethod
    def parse(source: Dict[str, List[Dict[str, str]]]) -> "Archive":
        """Get an Archive from a dict."""

        return from_dict(Archive, source)


class ArchiveProtocol(serde.Protocol):
    """Custom protocol for serializing Archives."""

    def load(self, file: BinaryIO) -> Archive:
        return Archive.parse(json.load(file))

    def dumps(self, value: Archive) -> bytes:
        return json.dumps(value.dict()).encode()


@dataclass(frozen=True)
class Accessor:
    """Artifact handler for sharded artifacts."""

    #: Archive the accessor is reading from
    archive: Archive
    #: Layer to read the archive with
    layer: "Layer"

    @overload
    def __getitem__(self, key: int) -> Any:
        ...

    @overload
    def __getitem__(self, key: slice) -> List[Any]:
        ...

    def __getitem__(self, key: Union[int, slice]) -> Any:
        datastore = self.layer.flow.configuration.datastore

        # Directly access an index
        if isinstance(key, int):
            if key >= len(self.archive):
                raise IndexError

            return datastore.read_artifact(layer=self.layer, archive=Archive(artifacts=[self.archive.artifacts[key]]))

        # Slicing for multiple splits
        elif isinstance(key, slice):
            return [
                datastore.read_artifact(layer=self.layer, archive=Archive(artifacts=[artifact]))
                for artifact in self.archive.artifacts[key]
            ]

        else:
            raise TypeError(f"{type(key)} is not a valid key type for Accessor.__getitem__")

    def __iter__(self) -> Generator[Any, None, None]:
        for artifact in self.archive.artifacts:
            yield self.layer.flow.configuration.datastore.read_artifact(
                layer=self.layer, archive=Archive(artifacts=[artifact])
            )

    def __len__(self) -> int:
        return len(self.archive)


@dataclass(frozen=True)
class DataStore:
    """Base flow datastore."""

    #: URI root of the datastore
    root: str
    #: Internal datastore cache
    cache: Dict[str, Any] = field(default_factory=dict)
    #: Custom serde protocols for reading/writing artifacts
    protocols: Dict[str, serde.Protocol] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if not self.root.endswith(("://", ":///")):
            object.__setattr__(self, "root", self.root.rstrip("/"))

        self.protocols[ArchiveProtocol.dtype] = ArchiveProtocol()
        self.protocols[RecordProtocol.dtype] = RecordProtocol()

    def uri(self, *, path: str) -> str:
        """Given a path, generate a URI in the datastore.

        Args:
            path: Path relative to the datastore root.

        Returns:
            URI to a location in the datastore.
        """

        return fs.join(self.root, path)

    def exists(self, *, path: str) -> bool:
        """Check if a file exists in the datastore.

        Args:
            path: Path from the datastore root to the file.

        Returns:
            True if the file exists else False
        """

        return fs.exists(uri=self.uri(path=path))

    def protocol(self, *dtypes: type) -> Callable[[Type[serde.ProtocolType]], Type[serde.ProtocolType]]:
        """Register a custom serde protocol for a type.

        Usage::

            @datastore.protocol(pd.DataFrame)
            def DataFrameProtocol(serde.Protocol):
                ...
        """

        def decorator(protocol: Type[serde.ProtocolType]) -> Type[serde.ProtocolType]:
            for dtype in dtypes:
                self.protocols[f"{dtype.__module__}.{dtype.__name__}"] = protocol()
            return protocol

        return decorator

    def read_archive(self, *, layer: Layer, index: int, name: str, cache: bool = False) -> Archive:
        """Read an archive from the laminar datastore.

        Args:
            layer: Layer being read from.
            index: Layer index being read from.
            name: Name of the artifact the archive is for.
            cache: Read the archive from the cache.

        Returns:
            Archive: Archive of the requested artifact.
        """

        archive: Archive = self._read(
            uri=self.uri(path=Archive.path(layer=layer, index=index, name=name, cache=cache)),
            dtype=ArchiveProtocol.dtype,
        )
        return archive

    def read_artifact(self, *, layer: Layer, archive: Archive) -> Any:
        """Read an artifact form the laminar datastore.

        Args:
            layer (Layer): Layer being read from.
            archive (Archive): Archive to reference the artifact from.

        Returns:
            Any: Artifact value.
        """

        # Read the artifact value
        if len(archive) == 1:
            artifact = archive.artifacts[0]
            return self._read(uri=self.uri(path=artifact.path(layer=layer)), dtype=artifact.dtype)

        # Create an accessor for the artifacts
        else:
            return Accessor(archive=archive, layer=layer)

    def read_record(self, *, layer: Layer) -> Record:
        """Read a layer record from the laminar datastore.

        Args:
            layer: Layer to get the record for.

        Returns:
            Layer record
        """

        record: Record = self._read(uri=self.uri(path=Record.path(layer=layer)), dtype=RecordProtocol.dtype)
        return record

    def _read(self, *, uri: str, dtype: str) -> Any:
        return self.protocols.get(dtype, DEFAULT_SERDE).read(uri)

    def read(self, *, layer: Layer, index: int, name: str) -> Any:
        """Read from the laminar datastore.

        Args:
            layer: Layer being read from.
            index: Layer index being read from.
            name: Name of the artifact being read.

        Returns:
            Any: Value of the artifact.
        """

        return self.read_artifact(layer=layer, archive=self.read_archive(layer=layer, index=index, name=name))

    def write_archive(self, *, layer: Layer, name: str, artifacts: List[Artifact], cache: bool = False) -> Archive:
        """Write an archive to the laminar datastore.

        Args:
            layer: Layer being written.
            name: Name of the archive being written.
            artifacts: Artfacts associated with the archive.
            cache: Write the archive to the cache.

        Returns:
            Archive metadata written to the laminar datastore.
        """

        archive = Archive(artifacts=artifacts)
        self._write(
            value=archive,
            uri=self.uri(path=archive.path(layer=layer, index=unwrap(layer.index, 0), name=name, cache=cache)),
            dtype=ArchiveProtocol.dtype,
        )
        return archive

    def write_artifact(self, *, layer: Layer, value: Any) -> Artifact:
        """Write an arifact to the laminar datastore.

        Args:
            layer: Layer being written.
            value: Value of the artifact to write.

        Returns:
            Artifact metadata written to the laminar datastore.
        """

        serializer = self.protocols.get(type(value).__name__, DEFAULT_SERDE)
        artifact = Artifact(
            dtype=serde.dtype(type(value)), hexdigest=hashlib.sha256(serializer.dumps(value)).hexdigest()
        )
        self._write(value=value, uri=self.uri(path=artifact.path(layer=layer)), dtype=artifact.dtype)

        return artifact

    def write_record(self, *, layer: Layer, record: Record) -> None:
        """Write a layer record to the laminar datastore.

        Args:
            layer: Layer the record is for.
            record: Record to write.
        """

        self._write(value=record, uri=self.uri(path=record.path(layer=layer)), dtype=RecordProtocol.dtype)

    def _write(self, *, value: Any, uri: str, dtype: str) -> None:
        self.protocols.get(dtype, DEFAULT_SERDE).write(value, uri)

    def write(self, *, layer: Layer, name: str, values: Iterable[Any]) -> None:
        """Write to the laminar datastore.

        Args:
            layer: Layer being written to.
            name: Name of the artifact being written.
            values : Artifact values to store.
        """

        artifacts = [self.write_artifact(layer=layer, value=value) for value in values]
        self.write_archive(layer=layer, name=name, artifacts=artifacts)

    def list_executions(self, *, flow: Flow) -> List[str]:
        """List all executions.

        Args:
            flow: Flow to list executions for.

        Returns:
            All executions.
        """

        return sorted(set(self._list(prefix=self.uri(path=fs.join(flow.name, "archives")), group="execution")))

    def list_layers(self, *, execution: Execution) -> List[str]:
        """List all layers in an execution.

        Args:
            execution: Execution to list layers for.

        Returns:
            All layers.
        """

        return sorted(
            set(
                self._list(
                    prefix=self.uri(path=fs.join(execution.flow.name, "archives", unwrap(execution.id))), group="layer"
                )
            )
        )

    def list_artifacts(self, *, layer: Layer) -> List[str]:
        """List all artifacts in a layer execution.

        Args:
            layer: Layer to list artifacts for.

        Returns:
            All artifacts.
        """

        return sorted(
            set(
                self._list(
                    prefix=self.uri(
                        path=fs.join(layer.flow.name, "archives", unwrap(layer.flow.execution.id), layer.name, "0")
                    ),
                    group="artifact",
                )
            )
        )

    def _list(self, *, prefix: str, group: str) -> Iterable[str]:
        raise NotImplementedError


@dataclass(frozen=True)
class Local(DataStore):
    """Store the laminar workspace on the local filesystem.

    Usage::

        Flow(datastore=Local())
    """

    root: str = str(Path.cwd() / ".laminar")

    def _list(self, *, prefix: str, group: str) -> Iterable[str]:
        for path in map(str, Path(prefix).glob("*")):
            match = ARCHIVE_PATTERN.match(path)
            if match is not None:
                yield match.group(group)


@dataclass(frozen=True)
class Memory(DataStore):
    """Store the laminar workspace in memory.

    Usage::

        Flow(datastore=Memory())
    """

    root: str = "memory:///"

    def exists(self, *, path: str) -> bool:
        return self.uri(path=path) in self.cache

    def _read(self, *, uri: str, dtype: str) -> Any:
        return self.cache[uri]

    def _write(self, *, value: Any, uri: str, dtype: str) -> None:
        self.cache[uri] = value

    def _list(self, *, prefix: str, group: str) -> Iterable[str]:
        for path in self.cache:
            if path.startswith(prefix):
                match = ARCHIVE_PATTERN.match(path)
                if match is not None:
                    yield match.group(group)


class AWS:
    @dataclass(frozen=True)
    class S3(DataStore):
        """Store the laminar workspace in AWS S3.

        Usage::

            Flow(datastore=AWS.S3())
        """

        def _list(self, *, prefix: str, group: str) -> Iterable[str]:
            parts = fs.parse_uri(prefix)

            s3 = boto3.resource("s3")
            bucket = s3.Bucket(name=parts.bucket)
            for response in bucket.objects.filter(Prefix=parts.key_id):
                match = ARCHIVE_PATTERN.match(response.key)
                if match is not None:
                    yield match.group(group)



"""Access for execution specific environment variables."""

from typing import Optional

from pydantic import BaseSettings


class Current:
    """Get information about the current execution environment.

    Usage::

        from laminar.settings import current

        current.execution.id
        current.flow.name
        current.layer.attempt
        current.layer.index
        current.layer.name
        currnet.layer.splits
    """

    class Execution(BaseSettings):
        class Config:
            env_prefix = "LAMINAR_EXECUTION_"

        #: ID of the current execution
        id: Optional[str] = None
        #: True if execution is being retried, else False
        retry: bool = False

    class Flow(BaseSettings):
        class Config:
            env_prefix = "LAMINAR_FLOW_"

        #: Name of the currently running flow
        name: Optional[str] = None

    class Layer(BaseSettings):
        class Config:
            env_prefix = "LAMINAR_LAYER_"

        #: Current layer attempt
        attempt: Optional[int] = None
        #: Index of the layer the layer splits
        index: Optional[int] = None
        #: Name of the currently running layer
        name: Optional[str] = None
        #: Number of splits in the layer
        splits: Optional[int] = None

    @property
    def execution(self) -> Execution:
        return self.Execution()

    @property
    def flow(self) -> Flow:
        return self.Flow()

    @property
    def layer(self) -> Layer:
        return self.Layer()


current = Current()



"""Test foreach flows."""

from typing import List, Tuple

import pytest

from laminar import Flow, Layer
from laminar.configurations import datastores, executors
from laminar.configurations.layers import ForEach, Parameter


class SingleGridFlow(Flow):
    ...


@SingleGridFlow.register()
class A(Layer):
    foo: List[int]
    bar: List[str]

    def __call__(self) -> None:
        self.shard(foo=[1, 2, 3])
        self.shard(bar=["a", "b"])


@SingleGridFlow.register(
    foreach=ForEach(parameters=[Parameter(layer=A, attribute="foo"), Parameter(layer=A, attribute="bar")])
)
class B(Layer):
    result: List[Tuple[str, int]]

    def __call__(self, a: A) -> None:
        self.result = (a.bar, a.foo)  # type: ignore


@SingleGridFlow.register()
class C(Layer):
    def __call__(self, b: B) -> None:
        self.result = list(b.result)


@pytest.mark.flow
def test_grid() -> None:
    flow = SingleGridFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow()

    assert list(execution.layer(A).foo) == [1, 2, 3]
    assert list(execution.layer(A).bar) == ["a", "b"]
    assert list(execution.layer(B).result) == [("a", 1), ("b", 1), ("a", 2), ("b", 2), ("a", 3), ("b", 3)]
    assert execution.layer(C).result == [("a", 1), ("b", 1), ("a", 2), ("b", 2), ("a", 3), ("b", 3)]


class MultiGridFlow(Flow):
    ...


@MultiGridFlow.register()
class A12(Layer):
    foo: List[int]

    def __call__(self) -> None:
        self.shard(foo=[1, 2, 3])


@MultiGridFlow.register()
class A22(Layer):
    bar: List[str]

    def __call__(self) -> None:
        self.shard(bar=["a", "b"])


@MultiGridFlow.register(
    foreach=ForEach(parameters=[Parameter(layer=A12, attribute="foo"), Parameter(layer=A22, attribute="bar")])
)
class B2(Layer):
    result: List[Tuple[str, int]]

    def __call__(self, a12: A12, a22: A22) -> None:
        self.result = (a22.bar, a12.foo)  # type: ignore


@MultiGridFlow.register()
class C2(Layer):
    def __call__(self, b: B2) -> None:
        self.result = list(b.result)


@pytest.mark.flow
def test_double_grid() -> None:
    flow2 = MultiGridFlow(datastore=datastores.Memory(), executor=executors.Thread())
    execution = flow2()

    assert list(execution.layer(A12).foo) == [1, 2, 3]
    assert list(execution.layer(A22).bar) == ["a", "b"]
    assert list(execution.layer(B2).result) == [("a", 1), ("b", 1), ("a", 2), ("b", 2), ("a", 3), ("b", 3)]
    assert execution.layer(C2).result == [("a", 1), ("b", 1), ("a", 2), ("b", 2), ("a", 3), ("b", 3)]



"""Shared laminar exceptions."""


class ExecutionError(Exception):
    """An error that occured during execution."""


class FlowError(Exception):
    """An error that occured during Flow definition."""


class LayerError(Exception):
    """An error that occured during Layer definition."""


class SchedulerError(Exception):
    """An error that occured during Flow scheduling."""



from typing import Any, Callable, Generator, Optional, TypeVar

T = TypeVar("T")

ATTRIBUTE = "annotation"

entry = "hook::entry"
execution = "hook::execution"
retry = "hook::retry"
submission = "hook::submission"


def annotate(hook: T, annotation: str) -> T:
    setattr(hook, ATTRIBUTE, annotation)
    return hook


def get(hook: Callable[..., Generator[Any, None, None]]) -> Optional[str]:
    return getattr(hook, ATTRIBUTE, None)
